import unittest
import time
from multiprocessing import Process
import os

import psutil
import requests
import pandas

from pyspark_gateway import server
from pyspark_gateway import PysparkGateway

class PysparkGatewayTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server = Process(target=server.run, kwargs={'debug': True, 'use_reloader': False})
        cls.server.start()
        cls.server.join(10) # needs some time to boot up the webserver

    @classmethod
    def tearDownClass(cls):
        cls.server.terminate()

        # just for testing so we don't have any
        # zombie procs because we're forking off forks
        # this isn't needed when run on its own
        cur_pid = os.getpid()

        for p in psutil.process_iter():
            if p.name() == 'python' and p.pid != cur_pid:
                p.terminate()

    def test_pyspark_gateway(self):
        pg = PysparkGateway()

        import pyspark
        from pyspark import SparkContext, SparkConf
        from pyspark.sql.context import SQLContext
        from pyspark.sql.functions import udf

        conf = SparkConf().set('spark.io.encryption.enabled', 'true')
        sc = SparkContext(gateway=pg.gateway, conf=conf)
        sqlContext = SQLContext.getOrCreate(sc)

        self.assertEqual(type(sc), SparkContext)

        df = sqlContext.createDataFrame([(1,2,'value 1')], ['id1', 'id2', 'val'])
        self.assertEqual(df.count(), 1)

        rows = df.collect()
        self.assertEqual(rows[0].id1, 1)

        pd = df.toPandas()
        self.assertEqual(type(pd), pandas.core.frame.DataFrame)

        data = [(1,2,'a'),(3,4,'b'),(5,6,'c')]
        df = sqlContext.createDataFrame(data, ['foo', 'bar', 'baz'])
        df.createOrReplaceTempView('foo_table')

        def squared(v):
            return v * v

        sqlContext.udf.register('squared', squared)

        squared_df = sqlContext.sql('select squared(foo) AS val from foo_table')
        rows = squared_df.collect()

        self.assertEqual(rows[2].val, '25')

        sc.stop()
        pg.gateway.shutdown()

if __name__ == '__main__':
    unittest.main()



import sys
import os
import atexit
import pkgutil

import requests
from py4j.java_gateway import JavaGateway, GatewayParameters

from pyspark_gateway.server import HTTP_PORT, GATEWAY_PORT

class PysparkGateway(object):
    host = None
    http_url = None
    gateway = None

    def __init__(self,
            host=os.environ.get('PYSPARK_GATEWAY_HOST', 'localhost'),
            http_port=os.environ.get('PYSPARK_GATEWAY_HTTP_PORT', HTTP_PORT)):

        self.host = host
        self.http_url = 'http://%s:%d' % (host, int(http_port))
        PysparkGateway.http_url = self.http_url
        PysparkGateway.host = self.host

        self.patch()
        self.check_version()
        self.start_gateway()

    @classmethod
    def open_tmp_tunnel(cls, port):
        r = requests.post(cls.http_url+'/tmp_tunnel', json={'port': port})
        return r.json()['port']

    def patch(self):
        path = os.path.dirname(os.path.realpath(__file__))+'/patch_files/java_gateway_patch.py'
        patch_file = open(path, 'r').read()

        pkg = pkgutil.get_loader('pyspark')

        path = pkg.get_filename().split('/')[:-1]
        path.append('java_gateway.py')
        path = '/'.join(path)

        if os.path.exists(path+'c'):
            os.remove(path+'c')

        original_file = open(path, 'r').read()

        with open(path, 'w') as f:
            f.write(patch_file)

        def put_back(data, path):
            with open(path, 'w') as f:
                f.write(data)

        atexit.register(put_back, original_file, path)

    def check_version(self):
        from pyspark_gateway.spark_version import spark_version, valid_spark_version
        from pyspark_gateway.version import __version__

        r = requests.get(self.http_url+'/spark_version')
        resp = r.json()

        server_major = resp['spark_major_version']
        server_minor = resp['spark_minor_version']
        server_patch = resp['spark_patch_version']

        client_major, client_minor, client_patch = spark_version()

        if not valid_spark_version():
            print('Pyspark Gateway requires Spark version >= 2.4')

            sys.exit(-1)
        elif 'pyspark_gateway_version' not in resp:
            print('Pyspark Gateway client version doesn\'t match server')
            print('No server version found.')

            sys.exit(-1)
        elif 'pyspark_gateway_version' in resp:
            if __version__ != resp['pyspark_gateway_version']:
                print('Pyspark Gateway client version doesn\'t match server')
                print('Server version: %s' % (resp['pyspark_gateway_version']))
                print('Client version: %s' % (__version__))

                sys.exit(-1)
        elif server_major != client_major or server_minor != client_minor:
            print('Spark server version: %s' % (resp['spark_version']))
            print('Spark client version: %s' % (spark_version))

            sys.exit('Spark version mismatch')

    def start_gateway(self):
        print('Starting local java gateway')

        r = requests.get(self.http_url+'/gateway')
        resp = r.json()

        params = GatewayParameters(
                address=self.host,
                port=GATEWAY_PORT,
                auth_token=resp['auth_token'],
                auto_convert=True)

        self.gateway = JavaGateway(gateway_parameters=params)



import flask
import time
import atexit
import signal
import os
import logging

logger = logging.getLogger()

from flask import Flask, request, jsonify
from flask.logging import default_handler

from pyspark_gateway.tunnel import TunnelProcess

HTTP_PORT = 25000
GATEWAY_PORT = 25001
TEMP_PORT = 25002

GATEWAY = None
TMP_PROC = None

app = Flask(__name__)

@app.route('/gateway')
def gateway():
    body = {'auth_token': GATEWAY.gateway_parameters.auth_token}

    return jsonify(body)

@app.route('/tmp_tunnel', methods=['POST'])
def temp_tunnel():
    global TMP_PROC, TEMP_PORT

    if TMP_PROC != None:
        TMP_PROC.proc.terminate()
        TMP_PROC.proc.join(1)

    req = request.json

    logger.info('Opening temporary tunnel from port %d to %d' % (req['port'], TEMP_PORT))

    TMP_PROC = TunnelProcess(TEMP_PORT, req['port'])
    TMP_PROC.proc.join(1)

    return jsonify({'port': TEMP_PORT})

@app.route('/spark_version', methods=['GET'])
def spark_version():
    from pyspark_gateway.spark_version import spark_version
    from pyspark_gateway.version import __version__

    major, minor, patch = spark_version()

    resp = {
        'spark_major_version': major,
        'spark_minor_version': minor,
        'spark_patch_version': patch,
        'pyspark_gateway_version': __version__
        }

    return jsonify(resp)

def run(*args, **kwargs):
    global GATEWAY

    if GATEWAY == None:
        from pyspark.java_gateway import launch_gateway

        GATEWAY = launch_gateway()
        TunnelProcess(GATEWAY_PORT, GATEWAY.gateway_parameters.port, keep_alive=True)

    if 'debug' not in kwargs or ('debug' in kwargs and kwargs['debug'] == False):
        app.logger.removeHandler(default_handler)
        app.logger = logger

        logger.info('Starting pyspark gateway server')

    if 'port' not in kwargs:
        kwargs['port'] = HTTP_PORT

    app.run(*args, **kwargs)

if __name__ == '__main__':
    run(debug=True, use_reloader=False, port=HTTP_PORT)



import sys

try:
    from pyspark.version import __version__
except:
    import findspark
    findspark.init()

    from pyspark.version import __version__

def spark_version():
    major, minor, patch = __version__.split('.')

    return int(major), int(minor), int(patch)

def valid_spark_version():
    major, minor, patch = spark_version()

    # requires spark >= 2.4
    if major >= 2 and minor >= 4:
        return True
    else:
        return False



#!/usr/bin/env python
import os
from distutils.core import setup

with open('README.rst') as fp:
    readme = fp.read()

setup(
    name='PysparkGateway',
    version='0.0.20',
    packages=[
        'pyspark_gateway',
        'pyspark_gateway.patch_files'],
    license='Apache 2.0',
    description='Connect Pyspark to remote clusters',
    long_description=readme,
    install_requires=[
        'requests>=2.20.0',
        'Flask~=1.0.2',
        'py4j~=0.10.7',
        'findspark~=1.3.0'
        ],
    extras_require={
        'dev': [
            'pandas',
            'psutil'
        ]
    },
    python_requires='>=2.7',
    scripts=['bin/pyspark-gateway'],
    url='https://github.com/abronte/PysparkGateway',
    author='Adam Bronte',
    author_email='adam@bronte.me',
)



import os
from multiprocessing import Process
import time
import threading
import socket
import signal
import logging

logger = logging.getLogger()

class Tunnel(object):
    bind_address = '127.0.0.1'
    remote_socket = None
    local_socket = None
    server_socket = None

    def __init__(self, src_port, dst_port, keep_alive=False):
        self.src_port = src_port
        self.dst_port = dst_port
        self.keep_alive = keep_alive

        signal.signal(signal.SIGTERM, self.signal_handler)

        self.server()

    def signal_handler(self, signum, frame):
        logger.info('Terminating tunnel from %d to %d' % (self.src_port, self.dst_port))

        if self.remote_socket:
            self.remote_socket.close()

        if self.local_socket:
            self.local_socket.close()

        if self.server_socket:
            self.server_socket.close()

        os._exit(0)

    def server(self):
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.server_socket.bind((self.bind_address, self.src_port))
        self.server_socket.listen(10)

        while True:
            self.local_socket, local_address = self.server_socket.accept()
            self.remote_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.remote_socket.connect((self.bind_address, self.dst_port))

            s = threading.Thread(target=self.handle, args=(self.remote_socket, self.local_socket))
            r = threading.Thread(target=self.handle, args=(self.local_socket, self.remote_socket))

            s.start()
            r.start()
        
    def handle(self, src, dst):
        try:
            while True:
                buffer = src.recv(1024)
                if len(buffer) == 0:
                    break

                dst.send(buffer)
        except:
            pass

        try:
            src.shutdown(socket.SHUT_RDWR)
            src.close()
            dst.shutdown(socket.SHUT_RDWR)
            dst.close()

            if self.keep_alive == False:
                self.signal_handler(None, None)
        except:
            pass

class TunnelProcess(object):
    proc = None

    def __init__(self, src_port, dst_port, timeout=None, keep_alive=False):
        self.proc = Process(target=Tunnel, args=(src_port, dst_port, keep_alive))
        self.proc.daemon = True
        self.proc.start()

        if timeout != None:
            t = threading.Thread(target=self.wait_and_kill, args=(self.proc, timeout))
            t.start()

    def wait_and_kill(self, proc, timeout):
        time.sleep(timeout)
        proc.terminate()
        proc.join(1)
